#
# PNlink.py
#  Módulo que conecta las Redes de Petri con el grafo
#  de alcanzabilidad / cubrimiento
#  
#  Desde el botón correspondiente del metamodelo de Red de Petri
#  (PetriNets.py), se incluirá este módulo que generará el
#  modelo en formalismo ReachGraph
#
# Autor: Jaime López
# Ultima modificación : 2003208
# $Id$
#

from Tkinter import *
import debug
import ReachGraph
import ReachGraph_MM
import ASG_ReachGraph
from ASG import *
from ATOM3 import *

def OpenReachGraph (AT3Dialog, ATOM3Instance, SemanticObject) :
    #ATOM3Instance.graph = AT3Dialog.parent.ASGroot
    l = PNlink()
    #print "AT3Dialog.parent = ", AT3Dialog.parent.ASGroot , "ATOM3Instance = ", ATOM3Instance.ASGroot
    currentState = l.pn_getCurrentState( AT3Dialog.parent.ASGroot )
    l.convert(AT3Dialog.parent, currentState, None, ATOM3Instance)
    for state in ATOM3Instance.ASGroot.listNodes["RGnode"]:
	    state.state.setValue(state.symName)

class PNlink :

	# Constructor
	def __init__(self) :
		self.iDebug = 0			# Imprimir informacin de debug
		self.iCubrimiento = 1		# 0=solo alcanzabilidad, 1=alcanzabilidad/cubrimiento
		self.GRID_SIZEX=230		# Tamaño de celda para posicionar nodos
		self.GRID_SIZEY=80		# Tamaño de celda para posicionar nodos
		self.iRec = 0			# Nivel de recursion actual

	#
	# Funciones de utilidad
	#
	
	# A partir de un diccionario, crea una lista
	# ordenada con sus valores
	# Permite ordenar por llave (por defecto) o por valor
	def _sort_dict(self, d, by_vals=0 ) :
		if by_vals : 
			items = [(v) for k,v in d.items()]
			items.sort()
		else :
			keys = d.keys()
			keys.sort()
			items = [ (d[k]) for k in keys ]
		return items
	
	# Convierte una lista en cadena
	def _list2str( self, l, sep = " ", print_brackets = 1 ) :
		s = ""
		if print_brackets : s = "[" + sep
		for i in l :
			s = s + str(i) + sep
		if print_brackets : s = s + "]"
		return s

	# Imprime con un cierto nivel de indentación
	def _print_indent ( self, s  ) :
		if self.iDebug == 1 :
			st = "%" + str(self.iRec) + "s " + s
			print st % ">"

	#
	# Functions of handling of the Petri Network
	#

	# It obtains the present, compound state by the values of tokens. Ej: [2 1 0]
	def pn_getCurrentState ( self, pn ) :
		aPlaces = pn.listNodes.get('PNPlace')
	 	aState = {} 
		for i in aPlaces :
			aState[ i.name.getValue() ] = i.tokens.getValue()
		return aState

	# It obtains the list of states accessible from the present state
	def pn_getCandidates ( self, pn, aState ) :

		# To obtain the list of qualified transitions
		aHabTrans = []
		aTrans = pn.listNodes.get('PNTransition')
		for oTrans in aTrans :
		 	bHabilitada = 1
			for t in oTrans.in_connections_ :
				n  = t.in_connections_[0]
				if aState[ n.name.getValue() ] < t.weight.getValue() :
					bHabilitada = 0
					break	
			if bHabilitada : 
				aHabTrans.append( oTrans )

		# To generate the new states possible (candidates) from the qualified transitions
		aCandidates = []
		for oTrans in aHabTrans :
			aNewState = aState.copy()
			# To eliminate tokens in the entrances
			for t in oTrans.in_connections_ :
				n = t.in_connections_[0]
				if aNewState[ n.name.getValue() ] == "w" :
				  	aNewState[ n.name.getValue() ] = "w"
				else :
					aNewState[ n.name.getValue() ] = aNewState[ n.name.getValue() ] - t.weight.getValue()
			# To add tokens in the exits
			for t in oTrans.out_connections_ :
				m = t.out_connections_[0]
				if aNewState[ m.name.getValue() ] == "w" :
				  	aNewState[ m.name.getValue() ] = "w"
				else :
					aNewState[ m.name.getValue() ] = aNewState[ m.name.getValue() ] + t.weight.getValue()
			# To add the new possible state
			aCandidates.append ( (aNewState, oTrans) )
		return aCandidates
	
	#
	# Functions of manipulation of the reachability graph
	#
	def rg_addNode ( self, atom_3, aCurrentState, oParentNode = None, iPos=1, transition = None ) :
		sNode = self._list2str( self._sort_dict( aCurrentState ) )
		n = atom_3.createNewRGnode( atom_3, 0, 0 )
		n.__aCurrentState__ = aCurrentState
		n.state.setValue( sNode )
		
		symName = "["
		first = 1
		for name in aCurrentState.keys():
			if aCurrentState[name] == 1:
				if not first: symName = symName + ", "
				symName = symName + name
				first = 0
			elif aCurrentState[name] != 0:
				if not first: symName = symName + ", "				
				symName = symName + name + "("+str(aCurrentState[name])+")"
				first = 0
		symName = symName + "]"
		n.symName = symName
		n.graphObject_.ModifyAttribute("state", symName)
		n.graphObject_.Move(iPos*self.GRID_SIZEX, self.iRec*self.GRID_SIZEY)
	
		self._print_indent( "Aadiendo " + sNode )
		if oParentNode != None :
		 	self.rg_connectPair( atom_3, oParentNode, n, transition )

		return n
	
	# It connects two nodes
	def rg_connectPair ( self, atom_3, oParentNode, oChildNode, transition ) :
	 	#c = atom_3.createNewRGtrans( atom_3, 0, 0)
		#c.out_connections_.append(oChildNode)
		#c.in_connections_.append(oParentNode)
		#oChildNode.in_connections_.append(c)
		#oParentNode.out_connections_.append(c)
		self._print_indent( "Connect "+oParentNode.state.getValue()+" with "+oChildNode.state.getValue() )
		atom_3.drawConnections((oParentNode, oChildNode))
		atom_3.newCreatedObject.transition.setValue(transition.name.getValue())
		atom_3.newCreatedObject.graphObject_.ModifyAttribute("transition", transition.name.getValue())

	# Looks for a certain state in the graph
	def rg_findState ( self, rg, state ) :
		s = self._list2str( self._sort_dict(state) )

		# Get the list of all nodes currently present
		nodes = rg.listNodes.get('RGnode')

		# Searching through each node at a time
		for n in nodes :
			# We have found an identical state so we return it right away
			if n.state.getValue() == s :
				return n

		# As the state is new, we check for dominance with its parent state from the initial state
		for n in nodes :
			isDominantPos = self.rg_isDominant( n, state)
			if isDominantPos != []:
				name = state.keys()
				name.sort()
				for pos in isDominantPos:
					state[name[pos]] = "w"
				s = self._list2str(self._sort_dict(state))
				break
		return 0

	# Determines if a node dominates another one (x over y)
	def rg_isDominant( self, y, x):
		iDomPos = []
		i = 0
		_aCurrentState = self._sort_dict(y.__aCurrentState__)
		_x = self._sort_dict(x)
		for v in _aCurrentState:
			if _x[i] == "w" and v!= "w": 
				iDomPos = iDomPos+[i]
			else:
				if v == "w" and _x[i] != "w":
					return []
				elif v != "w" and _x[i] != "w":
					if int(_x[i]) > int(v):
						iDomPos = iDomPos+[i]
					elif int(_x[i]) < int(v):
						return []
			i = i + 1
		return iDomPos

	#
	# MAIN ROUTINE
	# It turns the network of petri a graph of reachability of recursive form
	#
	def convert ( self, pn_atom3, aCurrentState=None, oParentNode=None, rg_atom3 = None, iPos = 1, transition = None) :

		# To obtain the Petri Network
		pn = pn_atom3.ASGroot

		# To obtain the reachability graph in its present state if it does not exist we created it
		if rg_atom3 == None :
			# New window of ATOM3
			rg_atom3 = ATOM3(Tk(), 'ReachGraph', 0, 1)
			aCurrentState = self.pn_getCurrentState( pn )
		else :
			self.iRec = self.iRec + 1
		rg = rg_atom3.ASGroot
	
		self._print_indent("---------------------")
		if aCurrentState != None :
		  self._print_indent("State=" + self._list2str(self._sort_dict(aCurrentState)))

		# To obtain the list of new states candidates, to start off of the active transitions
	        aNewStates = self.pn_getCandidates ( pn, aCurrentState )

		# To add the present node
		oCurrentNode = self.rg_addNode( rg_atom3, aCurrentState, oParentNode, iPos, transition )

		# For each one of these states, we advanced the network and recursiva call
		visited = 0
		iChildPos=0
		for s, t in aNewStates:
			iChildPos=iChildPos+1

			# If the state has already been visited, we break the recursion
			# This step includes checking for dominance of nodes and will assign the
			# w character if necessary to an unbounded network
			visited = self.rg_findState ( rg, s )
			if visited == 0 :
				# When the state has not been visited we recurse to look at all of its new states
	        		visited = self.convert( pn_atom3, s, oCurrentNode, rg_atom3, iChildPos, t)
			# When the state has been found we connect it to its parents node
			else :
				self.rg_connectPair( rg_atom3, oCurrentNode, visited, t )

		self.iRec = self.iRec-1
		return visited

